package polyglot.ext.polyj.util;

import polyglot.types.*;
import polyglot.util.*;
import polyglot.ext.polyj.types.*;
import java.util.*;

/**
 * A utility class for translation of methods.  Return types
 * can change during translation of PolyJ to Java.  This class
 * can be instantiated on a particular method to query whether
 * its return type needs to be changed because of the translationg of the
 * methods it overrides and implements, and if so, what the new type should be.
 */
public class ReturnTypeMutator {

    protected MethodInstance method;
    protected Type origRet;
    protected Type transRet;
    protected boolean typeChanges;
    protected PolyJTypeSystem ts;

    /**
     * @param method The method whose return type is to be examined.
     */
    public ReturnTypeMutator(MethodInstance method) {
	this.method = method;	
	origRet = method.returnType();
	ts = (PolyJTypeSystem) method.typeSystem();
	
	// Collect all the methods that this method overrides or implements.
	List superMethods = method.overrides();
	superMethods.addAll(method.implemented());
	
	// Uninstantiate (via originalMethod) all those methods,
	// filter out the ones whose types don't change, and collect
	// the set of return types from what remains.
	Set changedTypes = new HashSet( CollectionUtil.map(superMethods, 
	    new Transformation() {
	        public Object transform(Object o) {
		    MethodInstance m = (MethodInstance) o;
		    MethodInstance origM = ts.originalMethod(m);
		    Type retType = origM.returnType();
		    if (transChangesType(retType)) {
			return transType(retType);
		    } else {
			return Transformation.NOTHING;
		    }
		}
	    } ));
	
	int c = changedTypes.size();
	if (c > 1) {
	    throw new InternalCompilerError("Method will have multiple "
	        + "return types in translation, making it ambiguous: "
		+ method);
	} else if (c < 1) {
	    typeChanges = false;
	    transRet = origRet;
	} else {
	    typeChanges = true;
	    transRet = (Type) CollectionUtil.firstOrElse(changedTypes, null);
	}
    }

    protected boolean transChangesType(Type t) {
	Type u = transType(t);
	return !u.equals(t);
    }

    //HACK: this could be done much better by querying t itself,
    // but the machinery for that isn't in place yet.
    protected Type transType(Type t) {
	if (t instanceof ParamType) {
	    return ts.Object();
	} else if (t.isArray() 	    
	    && t.toArray().ultimateBase() instanceof ParamType)
	{
	    int d = t.toArray().dims();
	    if (d == 1) {
		return ts.Object();
	    } else {		    
		return ts.arrayOf(ts.Object(), d-1);
	    }
	} else {
	    return t;
	}
    }
    
    /**
     * Return whether the return type of the method changes
     * because of translation.
     */
    public boolean typeChanges() {
	return typeChanges;
    }

    /**
     * Return the return type the method should be given 
     * in translation.
     */
    public Type translatedType() {
	return transRet;
    }

    /**
     * Return the type to which a call to the method should
     * be casted.
     */
    public Type castTo() {
	return origRet;
    }
    
}

