package polyglot.ext.polyj.visit;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.frontend.Compiler;
import polyglot.main.*;
import polyglot.types.Package;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import polyglot.ext.polyj.qq.*;
import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;

import java.io.*;
import java.util.*;

public class PolyJTranslator extends Translator
{
    protected TranslationTable tt;
    protected boolean haveDumpedDebugInfo = false;
    protected static Map mangleMap;

    protected TargetFactory tf;
    protected PolyJTypeSystem ts;
    protected ErrorQueue eq;
    protected int outputWidth;
    protected Collection outputFiles;

    public PolyJTranslator(Job job, PolyJTypeSystem ts, NodeFactory nf, TargetFactory tf) {
        super(job, ts, nf, tf);

        this.tf = tf;
        this.ts = ts;
        this.eq = job.compiler().errorQueue();
        this.outputWidth = job.compiler().outputWidth();
        this.outputFiles = job.compiler().outputFiles();

        tt = ts.translationTable();
    }

    public QQ qq() {
        return new QQ(job.extensionInfo());
    }
    
    public boolean translate(Node ast) {
        if (!haveDumpedDebugInfo){
            // FIXME: add a switch for this
            // ((PolyJTypeSystem)ts).debug_dumpTranslationTable();
            haveDumpedDebugInfo = true;
        }

        if (! super.translate(ast)) {
            return false;
        }

        try {
            makeAdapterClasses();
        } catch (IOException e) {
            eq.enqueue( ErrorInfo.IO_ERROR,
                "I/O error while translating: " + e.getMessage());

            return false;
        }

        return true;
    }

    protected void makeAdapterClasses() throws IOException {
        for (Iterator instantiations = tt.instantiations().iterator();
            instantiations.hasNext(); )
        {
            ClassType t = (ClassType) instantiations.next();
            InstantiationBinding adapter = ts.InstantiationBinding(t);
            if (adapter.needsAdapterClass()) {
                makeAdapterClass(adapter);
            }
        }
    }

    /**
     * Create the adapter interface for the given type and write
     * it to the appropriate file.
     */
    public void makeAdapterInterface(PolyJClassType clazz) {
        // Determine the package the interface belongs in.
        String packageName = packageName(clazz);

        // Generate a $token class for this package, if necessary.
        makeTokenClass(packageName);

        // Determine the name of the interface.
        String adapterInterfaceName = adapterInterfaceShortName(clazz);

        String superAdapterName = null;
        ClassType superType = clazz.superType().toClass();
        if (superType != null && ts.needsWhereObject(superType)) {
            superAdapterName 
                = adapterInterfaceFullName(ts.uninstantiate(superType));
        }

        try {
            File adapterIntFile = tf.outputFile(packageName,
                adapterInterfaceName, job.source());
            outputFiles.add(adapterIntFile.getPath());
            Writer ofw = tf.outputWriter(adapterIntFile);
            CodeWriter w = new CodeWriter( ofw, outputWidth );

            writeInterfacePre(packageName, adapterInterfaceName,
                superAdapterName, w);
            writeIntClinitCall(clazz, adapterInterfaceName, w);
            writePsigSignatures(clazz, w);
            writeParamBindingSignatures(clazz, w);
            writeAdapterBindingSignatures(clazz, w);
            writeStaticGetSignatures(clazz, w);
            writeStaticSetSignatures(clazz, w);
            writeArrayAccessSignatures(clazz, w);
            writeNewArraySignatures(clazz, w);
            writeInstanceofMethod(clazz, w);
            writeInterfacePost(w);
            w.flush();

        } catch (IOException e) {
            eq.enqueue( ErrorInfo.IO_ERROR,
                        "I/O error while translating: " + e.getMessage());
        }
    }

    protected void writeInterfacePre(String packageName,
        String adapterInterfaceName, String superAdapterName,
        CodeWriter w)
    {
        // write package header
        if (packageName != "") {
            w.write("package " + packageName + ";");
            w.newline(0);
            w.newline(0);
        }

        // write the interface header
        w.write("public abstract class "+adapterInterfaceName);
        if (superAdapterName != null) {
            w.write(" extends "+superAdapterName);
        }
        w.write(" {");
        w.newline(4);
        w.begin(0);
    }

    protected void writeInterfacePost(CodeWriter w) {
        w.end();
        w.newline(0);
        w.write("}");
    }

    /**
     * Write the static $instanceof method used to implement the
     * instanceof operator on polymorphic types.
     */
    protected void writeInstanceofMethod(PolyJClassType c, CodeWriter w) {
        int nParams = ts.allTypeParams(c).size();
        if (nParams < 1) {
            w.write("// No $instanceof method needed; this class has no type"
                + " variables in scope.");
            w.newline(0);
            return;
        }

        String fullName = fullNameForClass(c);
        String fullNameDollar = fullNameDollar(c);

        w.newline(0);
        w.write("public static boolean $instanceof(Object o");
        for (int count=1; count <= nParams; count++) {
            w.write( ", String $"+count );
        }
        w.write(") {");
        w.newline(4);

        w.begin(0);
        w.write("return (o instanceof " + fullName + ")");
        w.newline(4);
        w.write("&& ((" + fullName + ") o).$instanceof$$" 
		    + fullNameDollar + "($1");
        for (int count=2; count <= nParams; count++) {
            w.write(", $"+count);
        }
        w.write(");");
        w.end();

        w.newline(0);
        w.write("}");
        w.newline(0);
    }


    /**
     * Creates a token class in the specified directory.
     * If the class already exists, returns without making any changes
     *
     * A helper method for makeAdapterInterface().
     */
    private void makeTokenClass( String packageName )
    {
        try {
            File tokenFile = tf.outputFile(packageName, "$token", job.source());
            outputFiles.add(tokenFile.getPath());
            Writer ofw = tf.outputWriter(tokenFile);
            CodeWriter w = new CodeWriter( ofw, outputWidth );

            // write package header
            if ( packageName != "" ) {
                w.write( "package "+packageName+";" );
                w.newline(0);
            }
            w.write("public final class $token {");
            w.newline(4);

            w.begin(0);
            w.write("static final $token the_instance = new $token();");
            w.newline(0);
            w.write("private $token() {}");
            w.end();

            w.newline(0);
            w.write("}");
            w.newline(0);
            w.flush();
        } catch (IOException e) {
            eq.enqueue( ErrorInfo.IO_ERROR,
                        "I/O error while generating $token class: " + e.getMessage());
        }
    }

    /**
     * Create the adapter class for the given instantiation.
     */
    protected void makeAdapterClass(InstantiationBinding instantiation)
            throws IOException 
    {
        ClassType inst = instantiation.type();

        String fullName = instantiation.getAdapterClassFullName();
        String packageName = instantiation.getAdapterClassPackageName();
        String shortName = instantiation.getAdapterClassShortName();
        String adapterInterfaceName = adapterInterfaceFullName(inst);
        String adapterInterfaceShortName = adapterInterfaceShortName(inst);

        File adapterClassFile = tf.outputFile(packageName, shortName, job.source());
        outputFiles.add(adapterClassFile.getPath());
        Writer ofw = tf.outputWriter(adapterClassFile);
        CodeWriter w = new CodeWriter( ofw, outputWidth );

        writeAdapterPre(packageName, shortName, adapterInterfaceName,
            adapterInterfaceShortName, w);
        writeInstHandlers(inst, shortName, adapterInterfaceName,
            adapterInterfaceShortName, w);
        writeAdapterMethods(inst, w);
        writeStaticFields(inst, w);
        writeAdapterPost(w);
    
        w.flush();
    }

    /**
     * Write the preamble, or first few lines, of the adapter class --
     * its package, imports, and class declaration.
     */
    protected void writeAdapterPre(String packageName, String shortName,
            String adapterInterfaceName, String adapterInterfaceShortName,
            CodeWriter w) 
    {
        // write package header
        if (packageName != "") {
            w.write("package " + packageName + ";");
            w.newline(0);
            w.newline(0);
        }

        // import the adapter interface
        w.write("import " + adapterInterfaceName + ";");
        w.newline(0);
        w.newline(0);

        // write the class header
        w.write("public final class " + shortName);
        w.write(" extends ");
        w.write(adapterInterfaceShortName + " {");
        w.newline(4);
        w.begin(0);
    }

    /**
     * Finish writing the adapter by closing everything
     * opened in writeAdapterPre().
     */
    protected void writeAdapterPost(CodeWriter w) {
        w.end();
        w.newline(0);
        w.write( "}" );
    }

    /**
     * Write the psig, default, where, and adapter binding methods
     * for all the parameters of the type.
     */
    protected void writeAdapterMethods(ClassType inst, CodeWriter w) {
        while (!ts.equals(inst,ts.Object())) {
            w.write("// Adapter methods for " + inst);
            w.newline(0);
            w.newline(0);

            Map pm = getParamMap(inst);

            writePsigMethods(inst, pm, w);
            writeParamBindingMethods(inst, pm, w);
            writeArrayAccessMethods(inst, pm, w);
            writeNewArrayMethods(inst, pm, w);
            writeAdapterBindingMethods(inst, w);

            inst = (ClassType) inst.superType();
        }
    }

    /**
     * Write the constructor and factory methods for the
     * adapter class.
     */
    protected void writeInstHandlers(ClassType inst, String shortName,
            String adapterInterfaceName, String adapterInterfaceShortName,
            CodeWriter w) 
    {
        // write instantiation handlers
        w.write("private " + shortName + "() { }" );
        w.newline(0);
        w.write("private static " + adapterInterfaceShortName 
            + " the_instance = new " + shortName + "();");
        w.newline(0);
        w.write("public static " + adapterInterfaceShortName
             + " getInstance() { return the_instance; }" );
        w.newline(0);
  
        if (clinitNeeded(inst)) {
            w.write("static { " + adapterInterfaceShortName
                + ".$clinit(the_instance); }");
            w.newline(0);
        }
        
        w.newline(0);
    }
    
    public boolean clinitNeeded(ClassType curClass) {
        boolean isParametric = ts.needsWhereObject(curClass);
        boolean isInterface = curClass.flags().isInterface();
        boolean isInnerNonStatic = curClass.isInner() 
            && !curClass.flags().isStatic();
        return isParametric && !isInterface && !isInnerNonStatic;
    }

    protected void writeIntClinitCall(ClassType curClass, String adIntName,
        CodeWriter w) 
    {
        if (clinitNeeded(curClass)) {
            w.write("public static void $clinit(" + adIntName + " $W) {");
            w.newline(4);
            w.begin(0);
            w.write(fullNameForClass(curClass) + ".$clinit($W);");
            w.end();
            w.newline();
            w.write("}");
            w.newline();
        }
    }
    
    /**
     * Write the PSig methods for c.
     */
    protected void writePsigMethods(ClassType c, Map paramMap, CodeWriter w)
    {
        writePsigMethodsImpl(c, paramMap, w, true);
    }

    /**
     * Write the PSig signatures for c.
     */
    protected void writePsigSignatures(ClassType c, CodeWriter w)
    {
        writePsigMethodsImpl(c, null, w, false);
    }

    /**
     * If writeBodies is false, only the
     * signatures of the methods will be written, as for a Java abstract class.
     * If writeBodies is true, the method bodies for c and its supertypes
     * will also be written, as for a Java class.
     */
    protected void writePsigMethodsImpl(ClassType c, Map paramMap,
        CodeWriter w, boolean writeBodies)
    {
        for (Iterator i = ts.allTypeParams(c).iterator();
            i.hasNext(); )
        {
            ParamType param = (ParamType) i.next();
            Type arg = null;
            if (writeBodies) {
                arg = (Type) paramMap.get(param);
            }
            writePsigMethod(c, param, arg, writeBodies, w);
        }

        w.newline(0);
    }

    /**
     * Write the PSig method for param instantiated on arg inside c.
     * If writeBody is false, arg is ignored and may be null.
     */
    protected void writePsigMethod(ClassType c, ParamType param, Type arg,
	        boolean writeBody, CodeWriter w) 
    {
        String methodSig = ts.ParamSigBinding(param, c).name() + "()";

        w.write("public ");
		if (!writeBody) {
		    w.write("abstract ");
		}

		w.write("String " + methodSig );
        if (!writeBody) {
            w.write(";");
        } else {
            w.write(" {");
            w.newline(4);
            w.begin(0);
            w.write("return \"" + arg + "\";");
            w.end();
			w.newline(0);
            w.write("}");
        }
        w.newline(0);
    }

    /**
     * Write the default and where signatures for the parameters
     * of inst.
     */
    protected void writeParamBindingSignatures(ClassType inst, CodeWriter w)
    {
        for (Iterator params = ts.allTypeParams(inst).iterator();
            params.hasNext(); )
        {
            ParamType param = (ParamType) params.next();
            writeDefaultSignature(inst, param, w);
            writeWhereSignatures(inst, param, w);

        }
        w.newline(0);
    }


    /**
     * Write the default and where methods for the parameters
     * of polyType.
     */
    protected void writeParamBindingMethods(ClassType inst, Map paramMap,
        CodeWriter w)
    {
        for (Iterator params = ts.allTypeParams(inst).iterator();
            params.hasNext(); )
        {
            ParamType param = (ParamType)params.next();
            writeDefaultMethod(inst, param, paramMap, w);
            writeWhereMethods(inst, param, paramMap, w);
            w.newline(0);
        }
    }

    /**
     * Write the signature of the default method for param.
     */
    protected void writeDefaultSignature(ClassType c, ParamType param,
        CodeWriter w)
    {
        writeDefaultMethodImpl(c, param, null, false, w);
    }

    /**
     * Write the default method for param.
     */
    protected void writeDefaultMethod(ClassType c, ParamType param,
        Map paramMap, CodeWriter w)
    {
        writeDefaultMethodImpl(c, param, paramMap, true, w);
    }

    /**
     * Write the default value method for param.
     */
    protected void writeDefaultMethodImpl(ClassType c, ParamType param,
        Map paramMap, boolean writeBody, CodeWriter w)
    {
        DefaultBinding defaultBinding = ts.DefaultBinding(param, c);
        w.write("// Default value accessor for " + param);
        w.newline(0);
        w.write("public ");
        if (!writeBody) {
            w.write("abstract ");
        }
        w.write("java.lang.Object " + defaultBinding.name() + "()");
        if (writeBody) {
            w.write(" { return ");
            Type actual = (Type) paramMap.get(param);
            if (actual.isPrimitive()) {
                w.write(ts.primitiveTypeWrapperName(actual.toPrimitive())
                    + ".getDefault(); }");
            } else {
                w.write(" null; }");
            }
        } else {
            w.write(";");
        }
        w.newline(0);
        w.newline(0);
    }

    /**
     * Write the where signatures for param.
     */
    protected void writeWhereSignatures(ClassType c, ParamType param,
        CodeWriter w)
    {
        writeWhereMethodsImpl(c, param, null, w, false);
    }

    /**
     * Write the where methods for param.
     */
    protected void writeWhereMethods(ClassType c, ParamType param,
        Map paramMap, CodeWriter w)
    {
        writeWhereMethodsImpl(c, param, paramMap, w, true);
    }

    /**
     * Write the where methods for param.
     */
    protected void writeWhereMethodsImpl(ClassType c, ParamType param,
        Map paramMap, CodeWriter w, boolean writeBodies)
    {
        w.write("// Where methods for " + param);
        w.newline(0);
        writeWhereProcedures(param.methods(), c, param, paramMap, 
	    w, writeBodies);
	w.newline(0);

	w.write("// Where constructors for " + param);
	w.newline(0);
        writeWhereProcedures(param.constructors(), c, param, paramMap, 
	    w, writeBodies);

	w.newline(0);	
    }

    /**
     * Write the where methods for the procedures contained in
     * <code>procs</code>.
     */
    protected void writeWhereProcedures(List procs, ClassType c,
        ParamType param, Map paramMap, CodeWriter w, boolean writeBodies)
    {
	for (Iterator i = procs.iterator(); i.hasNext(); ) {
	    ProcedureInstance pi = (ProcedureInstance) i.next();
	    WhereBinding binding = ts.WhereBinding(param, pi);
            if (!binding.isInherited()) {
                writeWhereMethod(c, param, binding, paramMap, w, writeBodies);
            } 
        }
    }
    
    /**
     * Write the where method represented by binding.
     */
    protected void writeWhereMethod(ClassType c, ParamType param,
        WhereBinding binding, Map paramMap, CodeWriter w, boolean writeBody)
    {
        writeWhereBindingHeader(w, binding, writeBody);

        if (writeBody) {
            w.begin(0);
            boolean writeArgs = true;

            if (binding.isConstructor()) {
                writeArgs = writeWhereConstructorCall(c, param, binding,
                    paramMap, w);
            } else {
                writeArgs = writeWhereMethodCall(c, param, binding,
                    paramMap, w);
            }

            if (writeArgs) {
                writeWhereMethodArgs(c, param, binding, paramMap, w);
            }

            w.write(";");
            w.end();
            w.newline(0);
            w.write( "}" );
            w.newline(0);
        }
    }

    /**
     * Write the header of a where method.  If isAbstract is true,
     * ends the header with ";".  If it's false, ends with " {".
     */
    protected void writeWhereBindingHeader(CodeWriter w,
            WhereBinding binding, boolean writeBody)
    {
        Iterator paramTypeIter = binding.procedureInstance().
            formalTypes().iterator();
        String methodName = binding.name();

        w.write("public ");
        if (!writeBody) {
            w.write("abstract ");
        }

        if (binding.isConstructor()) {
            w.write("java.lang.Object ");
        } else {
            MethodInstance methConst = binding.methodInstance();
            w.write(translateType(methConst.returnType()) + " ");
        }

        w.write(methodName + "(");

        if (binding.isMethod()) {
            w.write("java.lang.Object $this");
            if (paramTypeIter.hasNext()) {
                w.write(", ");
            }
        }

        int paramNum = 0;
        while (paramTypeIter.hasNext()) {
            Type paramType = (Type) paramTypeIter.next();
            paramNum++;
            w.write(translateType(paramType) + " $" + paramNum);
            if (paramTypeIter.hasNext()) {
                w.write(", ");
            }
        }

        w.write(")");
        if (writeBody) {
            w.write(" {");
        } else {
            w.write(";");
        }
        w.newline(0);
    }

    /**
     * Write the constructor call that is part of the where method
     * represented by binding.
     * @return true iff the arguments to the call should also be written
     */
    protected boolean writeWhereConstructorCall(ClassType c, ParamType param,
        WhereBinding binding, Map paramMap, CodeWriter w)
    {
        ConstructorInstance ci = binding.constructorInstance();
        Type actualType = (Type) paramMap.get(param);
        if (actualType.isPrimitive()) {
            w.write("return new " + ts.primitiveTypeWrapperName(
                actualType.toPrimitive()));
        } else if (actualType.isArray()) {
            if (ci.formalTypes().size() == 1) {
                Type firstArgType = (Type)
                    ci.formalTypes().get(0);
                if (firstArgType.isInt()) {
                    // this is the Array(int) constructor
                    w.write("return new " + fullNameForType(
                        actualType.toArray().base()) + "[$1]");
                } else {
                    // this is the Array(Array) constructor
                    w.write("return $1.clone()");
                }
                return false;
            } else {
                throw new InternalCompilerError(ci.position(),
                    "Illegal array constructor");
            }
        } else {
            w.write("return new " + fullNameForType(actualType));
        }

        return true;
    }

    /**
     * Write the method call that is part of the where method
     * represented by binding.
     * @return true iff the arguments to the call should also be written
     */
    protected boolean writeWhereMethodCall(ClassType c, ParamType param,
        WhereBinding binding, Map paramMap, CodeWriter w)
    {
        MethodInstance methConst = binding.methodInstance();
        Type returnType = methConst.returnType();
        if ( ! (returnType.isPrimitive()
                 && returnType.toPrimitive().isVoid() ))
        {
            String actualReturnTypeName
                = getActualTypeName(returnType, paramMap);
            w.write("return (" + actualReturnTypeName + ")");
        }
        w.write("((" + getActualTypeName(param, paramMap) + ")$this)."
            + methConst.name());

        return true;
    }

    /**
     * Write the arguments to the where method represented by
     * binding.
     */
    protected void writeWhereMethodArgs(ClassType c, ParamType param,
        WhereBinding binding, Map paramMap, CodeWriter w)
    {
        w.write("(");
        ProcedureInstance constraint = binding.procedureInstance();

        int paramNum = 0;
        for (Iterator paramTypeIter = constraint.formalTypes().iterator();
            paramTypeIter.hasNext(); )
        {
            Type paramType = (Type)paramTypeIter.next();
            paramNum++;

            String actualParamTypeName
                = getActualTypeName(paramType, paramMap);

            w.write("(" + actualParamTypeName + ")$" + paramNum);
            if (paramTypeIter.hasNext()) {
                w.write(", ");
            }
        }

        w.write(")");
    }

    /**
     * Write the declarations, as for a Java interface, of the
     * methods used to create new arrays of a parameter type.
     */
    protected void writeNewArraySignatures(ClassType poly, CodeWriter w) {
        writeNewArrayMethodsImpl(poly, null, w, false);
    }

    /**
     * Write the definitions, as for a Java class, of the
     * methods used to create new arrays of a parameter type.
     */
    protected void writeNewArrayMethods(ClassType poly,
        Map paramMap, CodeWriter w)
    {
        writeNewArrayMethodsImpl(poly, paramMap, w, true);
    }

    /**
     * Write the methods used to create new arrays of a parameter type.
     * If writeBodies is true, write the bodies of these methods.
     * If it is false, write just their signatures, making them abstract,
     * and paramMap may be null.
     */
    protected void writeNewArrayMethodsImpl(ClassType poly,
        Map paramMap, CodeWriter w, boolean writeBodies)
    {
        for (Iterator params = ts.allTypeParams(poly).iterator();
            params.hasNext(); )
        {
            ParamType param = (ParamType) params.next();

            w.write("// New array methods for " + param);
            w.newline(0);

            for (Iterator dims = tt.newArrayMethods(param);
                dims.hasNext(); )
            {
                Integer nDims = (Integer) dims.next();
                writeNewArrayMethod(poly, paramMap, w, writeBodies,
                    param, nDims.intValue());
            }
        }

        w.newline(0);
    }

    /**
     * Write the method to create a dims-dimensional array of
     * param.
     */
    protected void writeNewArrayMethod(ClassType poly,
        Map paramMap, CodeWriter w, boolean writeBodies,
        ParamType param, int dims)
    {
        NewArrayBinding binding = ts.NewArrayBinding(param, dims, poly);
        String repType = binding.type();
        Type actual = null;
        String actualName = null;
        if (writeBodies) {
            actual = (Type) paramMap.get(param);
            actualName = fullNameForType(actual);
        }

        w.write("public ");
        if (!writeBodies) {
            w.write("abstract ");
        }
        w.write(repType + " " + binding.name() + "(");
        for (int i = 1; i <= dims; i++) {
            if (i != 1) {
                w.write(", ");
            }
            w.write("int $" + i);
        }
        w.write(")");

        if (writeBodies) {
            w.newline(4);
            w.write("{ return (" + repType + ") new " + actualName);
            for (int i = 1; i <= dims; i++) {
                if (i != 1) {
                    w.write(", ");
                }
                w.write("[$" + i + "]");
            }
            w.write("; }");
        } else {
            w.write(";");
            w.newline(0);
        }
        w.newline(0);
    }

    /**
     * Write the declarations, as for a Java interface, of the
     * methods used to access arrays of a parameter type.
     */
    protected void writeArrayAccessSignatures(ClassType poly, CodeWriter w) {
        writeArrayAccessMethodsImpl(poly, null, w, false);
    }

    /**
     * Write the definitions, as for a Java class, of the
     * methods used to access arrays of a parameter type.
     */
    protected void writeArrayAccessMethods(ClassType poly, Map paramMap,
            CodeWriter w) {
        writeArrayAccessMethodsImpl(poly, paramMap, w, true);
    }

    /**
     * Write the methods used to access arrays of a parameter type.
     * If writeBodies is true, write the bodies of these methods.
     * If it is false, write just their signatures, making them abstract,
     * and paramMap may be null.
     */
    protected void writeArrayAccessMethodsImpl(ClassType poly,
            Map paramMap, CodeWriter w, boolean writeBodies)
    {
        for (Iterator params = ts.allTypeParams(poly).iterator();
                params.hasNext(); ) {
            ParamType param = (ParamType) params.next();

            w.write("// Array access methods for " + param);
            w.newline(0);
            ArrayAccessBinding binding = ts.ArrayAccessBinding(param, poly);
            Type actual = null;
            String actualName = null;
            String flags = "public";
            if (writeBodies) {
                actual = (Type) paramMap.get(binding.param());
                actualName = getTypeName(actual);
            } else {
                flags += " abstract";
            }

            writeArrayLengthMethod(w, flags, binding, writeBodies,
                actualName);
            writeArrayGetMethod(w, flags, binding, writeBodies, actual,
                actualName);
            writeArraySetMethod(w, flags, binding, writeBodies, actual,
                actualName);

            w.newline(0);
        }
    }

    /**
     * Write the method used to get the length of an array of a
     * parameter type.
     */
    protected void writeArrayLengthMethod(CodeWriter w, String flags,
        ArrayAccessBinding binding, boolean writeBodies, String actualName)
    {
        w.write(flags + " int " + binding.length() + "(Object $array)");
        if (writeBodies) {
            w.newline(4);
            w.write("{ return ((" + actualName + "[])$array).length; }");
        } else {
            w.write(";");
        }
        w.newline(0);
    }

    /**
     * Write the method used to get an element of an array of a
     * parameter type.
     */
    protected void writeArrayGetMethod(CodeWriter w, String flags,
        ArrayAccessBinding binding, boolean writeBodies,
        Type actual, String actualName)
    {
        String getBody = null;
        if (writeBodies) {
            if (actual.isPrimitive()) {
                String wrapperName = ts.wrapperTypeString(actual.toPrimitive());
                getBody = "{ return new " + wrapperName + "((("
                    + actualName + "[])$array)[$index]); }";
            } else {
                getBody = "{ return ((" + actualName
                    + "[])$array)[$index]; }";
            }
        }

        w.write(flags + " Object " + binding.get()
            + "(Object $array, int $index)");
        if (writeBodies) {
            w.newline(4);
            w.write(getBody);
        } else {
            w.write(";");
        }
        w.newline(0);
    }

    /**
     * Write the method used to set an element of an array of a
     * parameter type.
     */
    protected void writeArraySetMethod(CodeWriter w, String flags,
        ArrayAccessBinding binding, boolean writeBodies,
        Type actual, String actualName)
    {
        String setBody = null;
        if (writeBodies) {
            if (actual.isPrimitive()) {
                String wrapperName = ts.wrapperTypeString(actual.toPrimitive());
                setBody = "{ ((" + actualName + "[])$array)[$index] = (("
                        + wrapperName
                        + ")$value).Unwrap(); return $value; }";
            } else {
                setBody = "{ ((" + actualName + "[])$array)[$index] = ("
                    + actualName + ")$value; return $value; }";
            }
        }
        w.write(flags + " Object " + binding.set()
            + "(Object $array, int $index, Object $value)");
        if (writeBodies) {
            w.newline(4);
            w.write(setBody);
        } else {
            w.write(";");
        }
        w.newline(0);
     }

    /**
     * Given an instantiated type, returns a map from
     * formal type parameters to actual type parameters.
     */
    protected Map getParamMap(ClassType c) {
        Map paramMap = new HashMap();
        List params = ts.allTypeParams(c);
        List args = ts.allTypeArgs(c);
        Iterator pi = params.iterator();
        Iterator ai = args.iterator();
        while (pi.hasNext() && ai.hasNext()) {
            ParamType param = (ParamType) pi.next();
            Type arg = (Type) ai.next();
            paramMap.put(param, arg);
        }
        return paramMap;
    }

    protected void writeAdapterBindingSignatures(ClassType inst,
        CodeWriter w)
    {
        writeAdapterBindingMethodsImpl(inst, w, false);
    }

    protected void writeAdapterBindingMethods(ClassType inst, 
        CodeWriter w) 
    {
        writeAdapterBindingMethodsImpl(inst, w, true);
    }

    protected void writeAdapterBindingMethodsImpl(ClassType inst,
        CodeWriter w, boolean writeBodies)
    {
        PolyJClassType poly = ts.uninstantiate(inst);

        w.write("// Adapter bindings for " + poly);
        w.newline(0);

        for (Iterator bindings = tt.getAdapterBindings(poly);
            bindings.hasNext(); )
        {
            AdapterBinding binding = (AdapterBinding) bindings.next();
            String adapterInterfaceName = adapterInterfaceFullName(
                binding.instantiatedType());
            String instanceAdapter = null;
            if (writeBodies) {
                ClassType instance = tt.getInstantiatedClass(inst, binding);
                InstantiationBinding ib = ts.InstantiationBinding(instance);
                instanceAdapter = ib.getAdapterClassFullName();
            }

            w.begin(0);
            w.write("public ");
            if (!writeBodies) {
                w.write("abstract ");
            }
            w.write(adapterInterfaceName + " " + binding.name() + "()");
            if (writeBodies) {
                w.allowBreak(4);
                w.write("{ return " + instanceAdapter + ".getInstance(); }");
            } else {
                w.write(";");
            }

            w.end();
            w.newline(0);
        }

        w.newline(0);
    }

    /**
     * Write the declarations and methods associated with
     * static parametric fields.
     */
    protected void writeStaticFields(ClassType inst, CodeWriter w) {
        inst = ts.uninstantiate(inst);
        writeStaticFieldDecls(inst, w);
        writeStaticGetMethods(inst, w);
        writeStaticSetMethods(inst, w);
    }

    /**
     * Return the static fields of <code>t</code>, excluding the
     * pseudo-field <code>class</code>.
     */
    protected static Iterator staticFields(ClassType t) {
        return new TransformingIterator(t.fields().iterator(), 
            new Transformation() {
                public Object transform(Object o) {
                    FieldInstance fi = (FieldInstance) o;
                    if (!fi.flags().isStatic() || fi.name().equals("class")) {
                        return Transformation.NOTHING;
                    } else {
                        return o;
                    }
                }
            });
    }
    
    /**
     * Write the declarations of static parametric fields.
     */
    protected void writeStaticFieldDecls(ClassType inst, CodeWriter w) {
        w.write( "// static parametric fields" );
        w.newline(0);

        for (Iterator fieldIt = staticFields(inst); fieldIt.hasNext();) {
            FieldInstance fi = (FieldInstance) fieldIt.next();
            w.write(fi.flags().clearFinal().translate());
            w.write(translateType(fi.type()));
            w.write(" " + fi.name() + ";");
            w.newline(0);
            if (fi.flags().isFinal()) {
                w.write("private static boolean isInit$" + fi.name() 
                    + " = false;");
                w.newline(0);
            }
        }
    }

    protected void writeStaticGetSignatures(ClassType inst, CodeWriter w) {
        writeStaticGetMethodsImpl(inst, w, false);
    }

    protected void writeStaticGetMethods(ClassType inst, CodeWriter w) {
        writeStaticGetMethodsImpl(inst, w, true);
    }

    /**
     * Write the methods used to access static parametric fields.
     */
    protected void writeStaticGetMethodsImpl(ClassType inst, CodeWriter w,
        boolean writeBodies)
    {
	w.write("// static get methods for " + inst);
        w.newline(0);

        //This is the package of the polymorphic class.  (Needed for tokens)
        String polyPackage = packageName(inst);

        for (Iterator i = staticFields(inst); i.hasNext(); ) {
            FieldInstance field = (FieldInstance) i.next();
            StaticFieldBinding binding = ts.StaticFieldBinding(inst, field);
        
            w.write("public ");
            if (!writeBodies) {
                w.write("abstract ");
            }
            w.write(translateType(field.type()));
            w.write(" " + binding.get() + "(");
            if (!field.flags().isPublic()) {
                w.write(packagePrefix(polyPackage) + "$token $t)");
                if (writeBodies) {
                    w.write(" {");
                    w.newline(4);
                    w.begin(0);
                    w.write("if ($t == null) throw new java.lang.SecurityException();");
                    w.newline(0);
                } else {
                    w.write(";");
                    w.newline(0);
                }
            } else {
		w.write(")");
		if (writeBodies) {
		    w.write(" {");
		    w.newline(4);
		    w.begin(0);
		} else {
		    w.write(";");
		    w.newline(0);
		}
	    }
            if (writeBodies) {
                w.write("return " + field.name() + ";");
                w.end();
                w.newline(0);
                w.write("}");
                w.newline(0);
            }
        }
    }

    protected void writeStaticSetSignatures(ClassType inst, CodeWriter w) {
        writeStaticSetMethodsImpl(inst, w, false);
    }

    protected void writeStaticSetMethods(ClassType inst, CodeWriter w) {
        writeStaticSetMethodsImpl(inst, w, true);
    }

    /**
     * Write the methods used to assign to static parametric fields.
     */
    protected void writeStaticSetMethodsImpl(ClassType inst, CodeWriter w,
        boolean writeBodies)
    {
	w.write("// static set methods for " + inst);
        w.newline(0);

        //This is the package of the polymorphic class.  (Needed for tokens)
        String polyPackage = packageName(inst);

        for (Iterator i = staticFields(inst); i.hasNext(); ) {
            FieldInstance field = (FieldInstance) i.next();
            StaticFieldBinding binding = ts.StaticFieldBinding(inst, field);
            String fieldTypeString = translateType(field.type());

            w.write("public ");
            if (!writeBodies) {
                w.write("abstract ");
            }
            w.write(fieldTypeString + " " + binding.set() + "(");
            w.write(fieldTypeString);
            w.write(" $val");
            if (!field.flags().isPublic()) {
                w.write(", " + packagePrefix(polyPackage) + "$token $t)");
                if (writeBodies) {
                    w.write(" {");
                    w.newline(4);
                    w.begin(0);
                    w.write("if ($t == null) throw new java.lang.SecurityException();");
                    w.newline(0);
                } else {
                    w.write(";");
                    w.newline(0);
                }
            } else {
		w.write(")");
		if (writeBodies) {
		    w.write(" {");
		    w.newline(4);
		    w.begin(0);
		} else {
		    w.write(";");
		    w.newline(0);
		}
	    }
            if (writeBodies) {
                if (field.flags().isFinal()) {
                    w.write("if (isInit$" + field.name() + ") {");
                    w.newline(4);
                    w.begin(0);
                    w.write("throw new java.lang.IllegalStateException();");
                    w.newline(0);
                    w.end();
                    w.write("} else {");
                    w.newline(4);
                    w.begin(0);
                    w.write("isInit$" + field.name() + " = true;");
                    w.newline(0);
                    w.write("return (" + field.name() + "=$val);");
                    w.newline(0);
                    w.end();
                    w.write("}");
                    w.newline(0);
                } else {
                    w.write("return (" + field.name() + "=$val);");
                }
                w.end();
                w.newline(0);
                w.write("}");
                w.newline(0);
            }
        }
    }

    /**
     * Given a type object and a hashtable mapping formal parameter type names
     * to actual parameter type names, this method returns the name of the
     * actual type (using the given hashtable to map formal parameter types to
     * actual parameter types as necessary).
     */
    private String getActualTypeName(Type type, Map paramMap ) {
        if (type instanceof ParamType) {
            ParamType paramType = (ParamType) type;
            Type instantiatedType = (Type) paramMap.get(paramType);
            if (instantiatedType.isPrimitive()) {
                return ts.wrapperTypeString(instantiatedType.toPrimitive());
            } else if (instantiatedType.isArray()) {
                return getActualTypeName(instantiatedType.toArray().base(), paramMap) + "[]";
            } else {
                return fullNameForType(instantiatedType);
            }
        } else if (type.isArray()) {
            return getActualTypeName(type.toArray().base(), paramMap) + "[]";
        } else {
            return fullNameForType(type);
        }
    }

    /**
     * Return the full name for the given type, translating
     * it if it is an InstType.
     */
    private String getTypeName(Type type) {
        if (type instanceof InstType) {
            return type.translate(null);
        } else if (type.isArray()) {
            return getTypeName(type.toArray().base()) + "[]";
        } else {
            return fullNameForType(type);
        }
    }

    /**
     * Helper method for makeAdapterInterface().
     */
    protected String translateType( Type type ) {
        if ( type instanceof ParamType ) {
            return "java.lang.Object";
        } else if ( type instanceof ArrayType ) {
            if (((ArrayType)type).base() instanceof ParamType)
                return "java.lang.Object";
            else
                return translateType( ((ArrayType)type).base() )+"[]";
        } else {
            return fullNameForType(type);
        }
    }

    /**
     * Write the declaration of the instanceof class.
     */
    protected void makeInstanceofClassDecl(CodeWriter w, PolyJClassType interType)
    {
        String interfaceName = adapterInterfaceShortName(interType);

        // write the interface header
        w.write( "public class "+interfaceName + " {" );
        w.newline(4);
        w.begin(4);
        writeInstanceofMethod(interType, w);
        w.end();
        w.newline(0);
        w.write( "}" );
        w.newline(0);
    }

    /**
     * Creates an "instanceof" class for the given parametric interface
     * (which contains a static method for instanceof tests)
     * in the specified CodeWriter.  This has the same name that
     * the adapter interface for the interface would have.
     *
     * Rationale:  we need a static $instanceof method for each polymorphic class
     *  and interface.  Since we can't put static methods in an interface, we
     *  need to create a separate class.
     */
    public void makeInstanceofClass(PolyJClassType interType)
    {
        String className = adapterInterfaceShortName(interType);
        String packageName = packageName(interType);

        try {
            File adapterClassFile = tf.outputFile(packageName, 
                className, job.source());
            outputFiles.add(adapterClassFile.getPath());
            Writer ofw = tf.outputWriter(adapterClassFile);
            CodeWriter w = new CodeWriter( ofw, outputWidth );

            // write package header
            if ( packageName != "" ) {
                w.write( "package "+packageName+";" );
                w.newline(0);
                w.newline(0);
            }

            makeInstanceofClassDecl(w, interType);

            w.flush();
        } catch (IOException e) {
            eq.enqueue( ErrorInfo.IO_ERROR,
                        "I/O error while translating: " + e.getMessage());
        }
    }

    /**
     * Translate the target of the call normally.
     */
    public static void translateTargetNormal(CodeWriter w, Translator tr,
            Expr call, Node target) {
        translateTargetNoDot(w, tr, call, target);
        if (target != null) {
            w.write( ".");
        }
    }

    /**
     * Translate the target of the call normally, but leave off the
     * trailing dot if there should be one.
     */
    static void translateTargetNoDot(CodeWriter w, Translator tr,
            Expr call, Node target) {
        if (target != null) {
            if (target instanceof Expr) {
                call.printSubExpr((Expr) target, w, tr);
            } else {
                target.del().translate(w, tr);
            }
        }
    }

    /**
     * Translate the arguments in a procedure call.
     * Does not do boxing or unboxing.
     * Does add a where object for static methods.
     */
    public static void translateArgs(CodeWriter w, Translator tr,
            List arguments, ProcedureInstance proc) {
        w.write("(");
        translateArgsNoParens(w, tr, arguments, proc);
        w.write(")");
    }

    /**
     * Translate the arguments in a procedure call.
     * Does not do boxing or unboxing.
     * Does add a where object for static methods.
     * Does not include the parentheses around the argument list.
     */
    public static void translateArgsNoParens(CodeWriter w, Translator tr,
            List arguments, ProcedureInstance proc) {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        w.begin(0);

        // A static method or a constructor of a polymorphic class should
        // have a where object passed as its first argument,
        if (isStaticPolyProc(proc, ts) || isPolyConst(proc, ts)) {
            writeWhereArg(w, tr, arguments);
        }

        for (Iterator i = arguments.iterator(); i.hasNext(); ) {
            Expr e = (Expr) i.next();
            e.del().translate(w, tr);
            if (i.hasNext()) {
                w.write(", ");
                w.allowBreak(0);
            }
        }
        w.end();
    }

    /**
     * Translate the arguments in a call to a method in
     * an instantiated type.  This means boxing of
     * primitives that were originally parameters.
     * <code>targetType</code> should be set for methods,
     * and for new object creation expressions,
     * but should be null for explicit constructor calls.
     */
    public static void translateInstArgs(CodeWriter w, Translator tr,
            PolyJTypeSystem ts, Context c,
            List arguments, ClassType targetType,
            ProcedureInstance myPi, ProcedureInstance origPi) {

        w.write("(");
        w.begin(0);

        if (myPi.flags().isStatic()) {
            // A static method of a parametrized type should
            // have a where object passed as its first argument.
            writeStaticInstWhereArg(w, tr, ts, c, targetType.toClass(),
                arguments);
        } else if (isPolyConst(myPi, ts) && targetType != null) {
            // A constructor is essentially a static method
            writeStaticInstWhereArg(w, tr, ts, c, targetType.toClass(),
                arguments);
        } else if (isPolyConst(myPi, ts)) {
            // An internal constructor call takes a simpler
            // where argument.
            writeWhereArg(w, tr, arguments);
        }

        Iterator origArgIt = origPi.formalTypes().iterator();
        Iterator myArgIt = myPi.formalTypes().iterator();

        for (Iterator i = arguments.iterator(); i.hasNext(); ) {
            Expr e = (Expr) i.next();
            Type originalType = (Type) origArgIt.next();
            Type actualType = (Type) myArgIt.next();

            if (ts.isParameter(originalType) && actualType.isPrimitive()) {
                PrimitiveType primitive = actualType.toPrimitive();
                String primitiveWrapperName =
                    ts.primitiveTypeWrapperName(primitive);
                w.write("new " + primitiveWrapperName + "(");
                e.del().translate(w, tr);
                w.write(")");
            } else {
                e.del().translate(w, tr);
            }

            if (i.hasNext()) {
                w.write(",");
                w.allowBreak(0);
            }
        }

        w.end();
        w.write( ")");
    }

    public String whereObjExpr(){
	PolyJTypeSystem ts = (PolyJTypeSystem) typeSystem();
	ClassType c = context().currentClass();
	boolean inCxtr = context.currentCode() instanceof ConstructorInstance;
	boolean whereIsAmbig = ts.inheritsWhereObject(c) 
	    && ts.hasOuterWhereObject(c) && !inCxtr;
	if (whereIsAmbig) {
	    return "super.$W";
	} else {
    	    return "$W";
	}
    }
    
    /**
     * Write an uncasted where object as the first argument.
     */
    static void writeWhereArg(CodeWriter w, Translator tr, List arguments) {
        PolyJTranslator pjtr = (PolyJTranslator) tr;
	w.write(pjtr.whereObjExpr());
        if (!arguments.isEmpty()) {
            w.write(", ");
            w.allowBreak(0);
        }
    }

    /**
     * Write the first argument to a static instantiated method,
     * which is a where object casted to the right type.
     */
    static void writeStaticInstWhereArg(CodeWriter w, Translator tr,
            PolyJTypeSystem ts, Context c, ClassType targetType,
            List methArgs) {

        List arguments = ts.allTypeArgs(targetType);
        Iterator i = arguments.iterator();
        boolean isRealTypes = true;
        while (isRealTypes && i.hasNext()) {
            isRealTypes = !(i.next() instanceof ParamType);
        }

        // If the class is not being instantiated on other
        // parameter types, get an instance of the appropriate
        // adapter class.  Otherwise, ask the adapter class of
        // the current (scoped) polymorphic class to get the
        // adapter class on our behalf.
        if (isRealTypes) {
            InstantiationBinding ib = ts.InstantiationBinding(targetType);
            w.write(ib.getAdapterClassFullName()+".getInstance()");
        } else {
            PolyJClassType p = (PolyJClassType) c.currentClassScope();
            AdapterBinding ab = ts.AdapterBinding(p, targetType);
	    PolyJTranslator pjtr = (PolyJTranslator) tr;
            w.write("((" 
	        + pjtr.adapterInterfaceFullName(c.currentClassScope()) 
		+ ")" + pjtr.whereObjExpr() + ")." + ab.name()+"()");
        }

        if (!methArgs.isEmpty()) {
            w.write(",");
        }
    }

    /**
     * Translate a call to a where method that replaces a constructor.
     */
    public static void translateWhereConstructor(CodeWriter w, Translator tr,
        PolyJTypeSystem ts, Context c, ParamType targetParam,
        ProcedureInstance proc, List arguments) 
    {
        translateWhereMethod(w, tr, ts, c, targetParam, null, null,
            proc, arguments);
    }

    /**
     * Translate a call to a where method.
     * If the where method is actually the translation of a constructor,
     * (that is, proc is a ConstructorInstance), call and target may be
     * null.
     */
    public static void translateWhereMethod(CodeWriter w, Translator tr,
        PolyJTypeSystem ts, Context c, ParamType targetParam, Expr call,
        Node target, ProcedureInstance proc, List arguments) 
    {
	PolyJTranslator pjtr = (PolyJTranslator) tr;
        TranslationTable tt = ts.translationTable();
        WhereBinding whereMethod = ts.WhereBinding(targetParam, proc);
        String n = PolyJTranslator.adapterInterfaceFullName(
            c.currentClassScope());

        w.write("(("+ n +")" + pjtr.whereObjExpr() + ")." 
	    + whereMethod.name() + "(");

        w.begin(0);
        if (proc instanceof MethodInstance) {
            translateTargetNoDot(w, tr, call, target);
            if (!arguments.isEmpty()) {
                w.write(", ");
            }
        }
        translateArgsNoParens(w, tr, arguments, proc);
        w.end();

        w.write( ")");
    }

    static boolean isStaticPolyProc(ProcedureInstance proc,
            PolyJTypeSystem ts) {
        return ts.isPolymorphicClass(proc.container())
            && proc.flags().isStatic();
    }

    static boolean isPolyConst(ProcedureInstance proc,
            PolyJTypeSystem ts) {
        return ts.needsWhereObject(proc.container())
            && proc instanceof ConstructorInstance;
    }

    /**
     * Yield the full name of the current lexically enclosing class in
     * the Context c.  Throws an InternalCompilerError if that class
     * is not named.
     */
    public static String fullNameForCurrentClass(Context c) {
        return fullNameForClass(c.currentClassScope());
    }

    /**
     * Yield the (short) name of the current lexically enclosing class in
     * the Context c.  Throws an InternalCompilerError if that class
     * is not named.
     */
    public static String nameForCurrentClass(Context c) {
        return nameForClass(c.currentClassScope());
    }

    /**
     * Yield the full name of c, with every occurance of
	 * '.' replaced by '$'.
     * Useful for making a legal Java identifier out of
     * the full name of c.
     */
    public static String fullNameDollar(ClassType c) {
        return fullNameForClass(c).replace('.', '$');
    }

    /**
     * Yield the full name of c.  Throws an InternalCompilerError if c
     * is not named.
     */
    public static String fullNameForClass(ClassType c) {
      //assume c can't be null.
      String n = c.fullName();
     /*   if (c instanceof NamedType) {
            n = ((NamedType) c).fullName();
     */
     	
	if (n==null)
	{
   	throw new InternalCompilerError("The class has no name");
        }
        return n;
    }

    /**
     * Yield the (short) name of c.  Throws an InternalCompilerError if c
     * is not named.
     */
    public static String nameForClass(ClassType c) {
        //assume c can't be null.
	String n = c.name();
        if (n==null) 
	{
            throw new InternalCompilerError("The class has no name");
        }
        return n;
    }

    /**
     * Yield the full name of t.  Throws an InternalCompilerError if t
     * is not named.
     */
    public static String fullNameForType(Type t) {
        String n = null;
        if (t instanceof Named) {
            n = ((Named) t).fullName();
        } else {
            throw new InternalCompilerError("The type has no name");
        }
        return n;
    }

    /**
     * Yield the (short) name of t.  Throws an InternalCompilerError if t
     * is not named.
     */
     
    public static String nameForType(Type t) {
        String n = null;
        if (t instanceof Named) {
            n = ((Named) t).name();
        } else {
            throw new InternalCompilerError("The type has no name");
        }
        return n;
    }
    
    /**
     * Return the name of the package that <code>c</code>
     * is in.  If that's the default package, returns
     * the empty string.  If that's "polyj.util", returns "polyj.util".
     */
    protected static String packageName(ClassType c) {
        Package package_ = c.package_();
        if (package_ == null) {
            return "";
        } else {
            return package_.fullName();
        }
    }

    /**
     * Return the prefix used to refer to the package that <code>c</code>
     * is in.  If that's the default package, returns
     * the empty string.  If that's "p", returns "p.".
     */
    protected static String packagePrefix(ClassType c) {
        return packagePrefix(packageName(c));
    }

    /**
     * Return the prefix used to refer to packageName.
     * If packageName if empty, returns the empty string.
     * Otherwise, returns packageName + ".".
     */
    protected static String packagePrefix(String packageName) {
        if (packageName == "") {
            return "";
        } else {
            return packageName + ".";
        }
    }

    /**
     * Returns the fully qualified name
     * of the adapter interface for <code>c</code>.
     */
    public static String adapterInterfaceFullName(ClassType c) {
        return packagePrefix(c) + adapterInterfaceShortName(c);
    }

    /**
     * Returns the short (unqualified, i.e., no package) name
     * of the adapter interface for <code>c</code>.
     * If c's name is <name>, then the adapter interface for
     * c is: <name>{"."/"$$"}$$$W.
     * For example, the adapter interface for List[T].Node
     * would be List$$Node$$$W.
     */
    public static String adapterInterfaceShortName(ClassType c) {
        String className = fullNameForClass(c);
        String packageName = packageName(c);
        if (!packageName.equals("")) {
            // Here we assume that the full name for a class
            // always starts with its package name plus a dot.
            // Then we strip off that name plus dot.
            className = className.substring(packageName.length() + 1);
        }
        return replace(className, '.', "$$") + "$$$W";
    }

    /**
     * Same effects as String.replace, except replaces oldChar
     * with an entire String.
     */
    protected static String replace(String src, char oldChar, String newString) {
        Map subst = new HashMap(1);
        subst.put(new Character(oldChar), newString);
        return replace(src, subst);
    }

    /**
     * Maps src to a new string in which each character c of
     * src is replaced by:
     * - c, if c is not a key in subst
     * - s, if subst{c} = s
     * @param subst map from Character to String.
     */
    protected static String replace(String src, Map subst) {
        int length = src.length();
        StringBuffer dest = new StringBuffer(length);
        for (int i = 0; i < length; i++) {
            Character c = new Character(src.charAt(i));
            if (subst.containsKey(c)) {
                dest.append(subst.get(c));
            } else {
                dest.append(c.charValue());
            }
        }
        return dest.toString();
    }

    /**
     * Returns the string representing the type c
     * instantiated on args.
     */
    public static String instTypeString(ClassType c, List args) {
        StringBuffer buffer = new StringBuffer();
        buffer.append(c.toString());

        if (!args.isEmpty()) {
            buffer.append("[");
            for (Iterator i = args.iterator(); i.hasNext(); ) {
                Type arg = (Type) i.next();
                buffer.append(arg.toString());
                if (i.hasNext()) {
                    buffer.append(",");
                }
            }
            buffer.append("]");
        }

        return buffer.toString();
    }

    public static String adapterMethodName(String baseName, ParamType param,
        ClassType enclosing, PolyJTypeSystem ts) 
    {
	return adapterMethodName(baseName, param, enclosing, ts, null);	  
    }
    
    public static String adapterMethodName(String baseName, ClassType enclosing, 
        PolyJTypeSystem ts, String postfix)
    {
        return adapterMethodName(baseName, null, enclosing, ts, postfix);
    }
    
    /**
     * Return the name of an adapter method.
     * @param baseName The part of the name that identifies the purpose
     *                 of the method, e.g., "psig", "a_new".
     * @param param The parameter that the method is associated with.
     *              May be null.
     * @param enclosing The class that the method will be in.
     * @param postfix An additional string to add to the end of the name.
     *                May be null.
     */
    public static String adapterMethodName(String baseName, ParamType param,
        ClassType enclosing, PolyJTypeSystem ts, String postfix)
    {
        PolyJClassType polyType = ts.uninstantiate(enclosing);

        String param_str = "";
        if (param != null) {
            param_str = param.toString();
        }
        
        String from = "";
        if (!polyType.params().contains(param)) {
            from = "$$from$" + polyType;
        }
        
	String end = "";
	if (postfix != null && !postfix.equals("")) {
	    end = "$$" + postfix;
	}
	
        return mangle("$" + baseName + "$$" + param_str + from + end);
    }

    static {
        mangleMap = new HashMap(7);
        mangleMap.put(new Character('.'), "$");
        mangleMap.put(new Character(','), "$_");
        mangleMap.put(new Character('['), "$1");
        mangleMap.put(new Character(']'), "$2");
        mangleMap.put(new Character('<'), "$4");
        mangleMap.put(new Character('>'), "$7");
        mangleMap.put(new Character('#'), "$8");
    }

    /**
     * Mangle s to make it a possibly legal Java type
     * identifier by replacing characters in it.
     */
    public static String mangle(String s) {
        return replace(s, mangleMap);
    }

}
